package ec;

import ec.EvolutionState;
import ec.Evolve;
import ec.util.MersenneTwisterFast;
import ec.util.Output;
import ec.util.Parameter;
import ec.util.ParameterDatabase;

import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;


/**
 * Custom launcher to replace ec.Evolve.main()
 * Separates termination procedure, and makes available job and seed information
 * @author      Michael Gircys <mg12vp@brocku.ca>
 * @version     23.1.0
 * @since       23.1.0
 */
public class Launcher extends Evolve 
{

	private static int job = 0;
	public static final int GetCurrentJob(){ return job; }
	private static int numJobs = 1;
	public static final int GetTotalJobs(){ return numJobs; }

	private static int[] seeds = null;
	public static final int[] GetSeeds(){ return seeds; }

	private static EvolutionState state = null;
	private static ParameterDatabase parameters = null;

	private static final DateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");


	// Largely copied from ec.Evolve, seed values extracted
	public static EvolutionState initialize(ParameterDatabase parameters, int randomSeedOffset)
	{
		return initialize(parameters, randomSeedOffset, buildOutput());
	}
	public static EvolutionState initialize(ParameterDatabase parameters, int randomSeedOffset, Output output)
	{
		// Should we muzzle stdout and stderr?
		if (parameters.getBoolean(new Parameter(P_SILENT), null, false))
		{
			output.getLog(0).silent = true;
			output.getLog(1).silent = true;
		}
		// output was already created for us.  
		//output.systemMessage(Version.message());
		
		// 2. set up thread values
		int breedthreads = Evolve.determineThreads(output, parameters, new Parameter(P_BREEDTHREADS));
		int evalthreads  = Evolve.determineThreads(output, parameters, new Parameter(P_EVALTHREADS));
		// at least one thread is automatic?  Seeds may need to be dynamic.
		boolean auto =
			V_THREADS_AUTO.equalsIgnoreCase(parameters.getString(new Parameter(P_BREEDTHREADS),null)) ||
			V_THREADS_AUTO.equalsIgnoreCase(parameters.getString(new Parameter(P_EVALTHREADS),null)) ; 
		
		// 3. create the Mersenne Twister random number generators,
		// one per thread
		MersenneTwisterFast[] random = new MersenneTwisterFast[breedthreads > evalthreads ? breedthreads : evalthreads];
		seeds = new int[random.length];         
		String seedMessage = "Seed: ";
		
		byte[] addrs = new byte[4];
		int addr = 0;
		try{
			addrs = java.net.InetAddress.getLocalHost().getAddress();
			addr = addrs[0]<<24 | addrs[1]<<16 | addrs[2]<<8 | addrs[3] ;
		}catch(java.net.UnknownHostException ex){}
		int time = (int)(System.nanoTime());
		// nanoTime is delta time from startup within java environment, may overlap.
		// On a cluster, for many runs executing concurrently, lets add some uniqueness, like IP.
		time += addr; 
		
		for (int x=0; x<random.length; x++)
		{
			seeds[x] = determineSeed(output, parameters, new Parameter(P_SEED).push(""+x), time+x ,random.length * randomSeedOffset, auto);
			for(int y=0;y<x;y++)
				if( seeds[x] == seeds[y] )
					output.fatal( P_SEED+"."+x+" ("+seeds[x]+") and "+P_SEED+"."+y+" ("+seeds[y]+") ought not be the same seed.", null, null ); 
			random[x] = Evolve.primeGenerator(new MersenneTwisterFast(seeds[x]));    // we prime the generator to be more sure of randomness.
			seedMessage = seedMessage + seeds[x] + " ";
		}
		
		// 4.  Start up the evolution
		// what evolution state to use?
		EvolutionState state = null;
		state = (EvolutionState)parameters.getInstanceForParameter( new Parameter(P_STATE), null, EvolutionState.class );
		state.parameters = parameters;
		state.random = random;
		state.output = output;
		state.evalthreads = evalthreads;
		state.breedthreads = breedthreads;
		state.randomSeedOffset = randomSeedOffset;
		output.systemMessage("Threads:  breed/" + breedthreads + " eval/" + evalthreads);
		output.systemMessage(seedMessage);
		
		return state;
	}
	
	// Largely copied from ec.Evolve, hard System.exit(0) removed.
	public static void EvolveMain(String[] args)
	{
		checkForHelp(args);
		
		// if we're loading from checkpoint, let's finish out the most recent job
		int currentJob = 0;
		state = possiblyRestoreFromCheckpoint(args);
		if (state != null)  // loaded from checkpoint
		{
			try
			{
				if (state.runtimeArguments == null)
					Output.initialError("Checkpoint completed from job started by foreign program (probably GUI).  Exiting...");
				args = state.runtimeArguments;                          // restore runtime arguments from checkpoint
				currentJob = ((Integer)(state.job[0])).intValue() + 1;  // extract next job number
			}
			catch (Exception e)
			{
				Output.initialError("EvolutionState's jobs variable is not set up properly.  Exiting...");
			}
			
			state.run(EvolutionState.C_STARTED_FROM_CHECKPOINT);
			cleanup(state);
		}
		
		parameters = loadParameterDatabase(args);
		if (currentJob == 0)  // no current job number yet
			currentJob = parameters.getIntWithDefault(new Parameter("current-job"), null, 0);
		if (currentJob < 0)
			Output.initialError("The 'current-job' parameter must be >= 0 (or not exist, which defaults to 0)");
		numJobs = parameters.getIntWithDefault(new Parameter("jobs"), null, 1);
		if (numJobs < 1)
			Output.initialError("The 'jobs' parameter must be >= 1 (or not exist, which defaults to 1)");
		int firstJob = currentJob;
				
		for(job = currentJob ; job < numJobs; job++)
		{
			// load the parameter database (reusing the very first if it exists)
			if(currentJob != firstJob)
				parameters = loadParameterDatabase(args);
						
			// Initialize the EvolutionState, then set its job variables
			state = initialize(parameters, job);               // pass in job# as the seed increment
			state.output.systemMessage("Job: " + job);
			state.job = new Object[1];                         // make the job argument storage
			state.job[0] = Integer.valueOf(job);               // stick the current job in our job storage
			state.runtimeArguments = args;                     // stick the runtime arguments in our storage
			if (numJobs > 1)                                   // only if iterating (so we can be backwards-compatible),
			{
				String jobFilePrefix = "job." + job + ".";
				state.output.setFilePrefix(jobFilePrefix);     // add a prefix for checkpoint/output files 
				state.checkpointPrefix = jobFilePrefix + state.checkpointPrefix;  // also set up checkpoint prefix
			}
			
			// now we let it go
			state.run(EvolutionState.C_STARTED_FRESH);
			cleanup(state);  // flush and close various streams, print out parameters if necessary
			
			Date dateStart = new Date();
			state.output.systemMessage( "Job " + String.valueOf(job) + " Completed  : " + dateFormat.format(dateStart) );
			state.output.flush();
		}
		
	}

	public static void Termination()
	{
		// Kill our other windows and things.
		System.exit(0);
	}

	// Wrapper for main launch.
	public static void main(String[] args)
	{	
		Date dateStart = new Date();
		System.out.println( "Launched at  : " + dateFormat.format(dateStart) );
		
		EvolveMain(args);
		
		Date dateEnd = new Date();
		System.out.println( "Completed at : " + dateFormat.format(dateEnd)   );
		
		if(parameters != null && state != null)
		{
			String prob = state.evaluator.p_problem.getClass().getCanonicalName();
			String conf = parameters.getLabel();
			System.out.println("Completed problem [" + prob + "] configured with [" + conf + "]");
		}
		
		Termination();
	}

}
